import { generateSignature } from '@zoltu/ethereum-abi-encoder'
import { keccak256 } from '@zoltu/ethereum-crypto'

interface AbiParameter {
	name: string,
	type: string,
	components?: Array<AbiParameter>
}

export interface AbiEventParameter extends AbiParameter {
	indexed: boolean,
}

interface AbiEntry {
	type: string
}

interface AbiFunction extends AbiEntry {
	name: string,
	type: 'function',
	constant: boolean,
	payable: boolean,
	stateMutability: 'pure' | 'view' | 'payable' | 'nonpayable',
	inputs: Array<AbiParameter>,
	outputs: Array<AbiParameter>,
}

export interface AbiEvent {
	name: string,
	type: 'event',
	inputs: Array<AbiEventParameter>,
	anonymous: boolean,
}

export interface EventDescription {
	name: string
	signature: string
	signatureHash: string
	parameters: Array<AbiEventParameter>
}

type Abi = Array<AbiEntry>

interface CompilerOutput {
	contracts: {
		[globalName: string]: {
			[contractName: string]: {
				abi: Abi
			}
		}
	}
}

export async function generateContractInterfaces(contractsOutput: CompilerOutput): Promise<string> {
	const contractInterfaces: Array<string> = []
	const eventDescriptions: Map<string, string> = new Map<string, string>()
	const eventInterfaces: Array<string> = []
	const eventTypes: Array<string> = []

	for (let globalName in contractsOutput.contracts) {
		for (let contractName in contractsOutput.contracts[globalName]) {
			const contractAbi: Abi = contractsOutput.contracts[globalName][contractName].abi
			if (contractAbi.length == 0) continue
			contractInterfaces.push(contractInterfaceTemplate(contractName, contractAbi))
			for (let abiEvent of contractAbi.filter(abiEntry => abiEntry.type === 'event').map(abiEntry => <AbiEvent>abiEntry).filter(abiEvent => abiEvent.name)) {
				const eventDescription = await eventDescriptionTemplate(abiEvent)
				eventDescriptions.set(eventDescription.substring(1, 65), eventDescription)
				const eventInterface = eventInterfaceTemplate(contractName, abiEvent)
				eventInterfaces.push(eventInterface)
				eventTypes.push(`${contractName}.${abiEvent.name}`)
			}
		}
	}

	return `// THIS FILE IS AUTOMATICALLY GENERATED BY \`generateContractInterfaces.ts\`. DO NOT EDIT BY HAND'

import { Bytes32, Bytes, Address, EncodableArray, EncodableTuple } from '@zoltu/ethereum-types'
import { EventDescription, DecodedEvent, ParameterDescription, decodeParameters, decodeEvent } from '@zoltu/ethereum-abi-encoder'

interface Log {
	readonly topics: ReadonlyArray<Bytes32>
	readonly data: Bytes
}
interface TransactionReceipt {
	readonly status: boolean
	readonly logs: Iterable<Log>
}

export const eventDescriptions: { [signatureHash: string]: EventDescription & {signature: string} } = {
${Array.of(...eventDescriptions.values()).map(x => `\t${x}`).join(',\n')}
}

${eventInterfaces.join('\n\n')}

export type Event = DecodedEvent${eventTypes.length !== 0 ? ' | ': ''}${eventTypes.join(' | ')}


export interface Dependencies {
	call(address: Address, methodSignature: string, methodParameters: EncodableArray, value: bigint): Promise<Uint8Array>
	submitTransaction(address: Address, methodSignature: string, methodParameters: EncodableArray, value: bigint): Promise<TransactionReceipt>
}


/**
 * By convention, pure/view methods have a \`_\` suffix on them indicating to the caller that the function will be executed locally and return the function's result.  payable/nonpayable functions have both a local version and a remote version (distinguished by the trailing \`_\`).  If the remote method is called, you will only get back a transaction hash which can be used to lookup the transaction receipt for success/failure (due to EVM limitations you will not get the function results back).
 */
export class Contract {
	protected constructor(protected readonly dependencies: Dependencies, public readonly address: Address) { }

	protected async localCall(methodSignature: string, outputParameterDescriptions: ReadonlyArray<ParameterDescription>, methodParameters: EncodableArray, attachedEth?: bigint): Promise<EncodableTuple> {
		const result = await this.dependencies.call(this.address, methodSignature, methodParameters, attachedEth || 0n)
		return decodeParameters(outputParameterDescriptions, result)
	}

	protected async remoteCall(methodSignature: string, parameters: EncodableArray, errorContext: { transactionName: string }, attachedEth?: bigint): Promise<Array<Event>> {
		const transactionReceipt = await this.dependencies.submitTransaction(this.address, methodSignature, parameters, attachedEth || 0n)
		if (!transactionReceipt.status) throw new Error(\`Remote call of \${errorContext.transactionName} failed: \${JSON.stringify(transactionReceipt)}\`)
		return this.decodeEvents(transactionReceipt.logs)
	}

	private decodeEvents(encodedEvents: Iterable<Log>): Array<Event> {
		const decodedEvents: Array<DecodedEvent> = []
		for (const encodedEvent of encodedEvents) {
			const decodedEvent = this.tryDecodeEvent(encodedEvent)
			if (decodedEvent) decodedEvents.push(decodedEvent)
		}
		return decodedEvents as Array<Event>
	}

	private tryDecodeEvent(encodedEvent: Log): DecodedEvent | null {
		const signatureHash = encodedEvent.topics[0]
		const eventDescription = eventDescriptions[signatureHash.toString()]
		if (!eventDescription) return null
		return decodeEvent(eventDescription, encodedEvent.topics, encodedEvent.data)
	}
}

${contractInterfaces.join('\n')}`
}

function contractInterfaceTemplate(contractName: string, contractAbi: Abi) {
	const contractMethods: Array<string> = []

	// FIXME: Add support for Solidity function overloads.  Right now overloaded functions are not supported, only the first one seen will servive addition into the following set.
	const seen: Set<string> = new Set()

	const contractFunctions: Array<AbiFunction> = contractAbi
		.filter(abiEntry => abiEntry.type == 'function')
		.map(abiFunction => <AbiFunction>abiFunction)

	for (let abiFunction of contractFunctions) {
		if (seen.has(abiFunction.name)) continue
		if (!abiFunction.constant) {
			contractMethods.push(remoteMethodTemplate(abiFunction, { contractName: contractName}))
		}
		contractMethods.push(localMethodTemplate(abiFunction, { contractName: contractName}))
		seen.add(abiFunction.name)
	}

	return `
export class ${contractName} extends Contract {
	public constructor(dependencies: Dependencies, address: Address) {
		super(dependencies, address)
	}

${contractMethods.join('\n\n')}
}
`
}

async function eventDescriptionTemplate(abiEvent: AbiEvent): Promise<string> {
	const signature = toSignature(abiEvent.name, abiEvent.inputs)
	const signatureHash = await keccak256.hash(new TextEncoder().encode(signature))
	const eventDescription = {
		type: 'event',
		name: abiEvent.name,
		signature: signature,
		inputs: abiEvent.inputs,
	}
	return `'${signatureHash.toString(16)}': ${JSON.stringify(eventDescription)}`
}

function eventInterfaceTemplate(contractName: string, abiEvent: AbiEvent): string {
	return `export namespace ${contractName} {
	export interface ${abiEvent.name} extends DecodedEvent {
		name: '${abiEvent.name}'
		parameters: {
${
	abiEvent.inputs
		.map(parameter => (parameter.indexed && parameter.type === 'string') ? { ...parameter, type: 'bytes32' } : parameter)
		.map(input => `\t\t\t${input.name}: ${toTsTypeString(input, false, { contractName: contractName, functionName: abiEvent.name })}`).join('\n')
}
		}
	}
}`
}

function remoteMethodTemplate(abiFunction: AbiFunction, errorContext: { contractName: string }) {
	const argNames: string = toArgNameString(abiFunction)
	const params: string = toParamsString(abiFunction, true, errorContext)
	const methodSignature = generateSignature(abiFunction)
	const separator = (abiFunction.inputs.length !== 0 && abiFunction.payable) ? ', ' : ''
	const attachedEthInputParameter = abiFunction.payable ? 'attachedEth?: bigint' : ''
	const attachedEthCallParameter = abiFunction.payable ? ', attachedEth' : ''
	return `	public ${abiFunction.name} = async (${params}${separator}${attachedEthInputParameter}): Promise<Array<Event>> => {
		const methodSignature = '${methodSignature}' as const
		const methodParameters = [${argNames}] as const
		return await this.remoteCall(methodSignature, methodParameters, { transactionName: '${abiFunction.name}' }${attachedEthCallParameter})
	}`
}

function localMethodTemplate(abiFunction: AbiFunction, errorContext: { contractName: string }) {
	const argNames: string = toArgNameString(abiFunction)
	const params: string = toParamsString(abiFunction, true, errorContext)
	const methodSignature = generateSignature(abiFunction)
	const separator = (abiFunction.inputs.length !== 0 && abiFunction.payable) ? ', ' : ''
	const attachedEthInputParameter = abiFunction.payable ? 'attachedEth?: bigint' : ''
	const attachedEthCallParameter = abiFunction.payable ? ', attachedEth' : ''
	const returnType: string = toTsReturnTypeString(abiFunction.outputs, { contractName: errorContext.contractName, functionName: abiFunction.name })
	const returnValue: string = (abiFunction.outputs.length === 1)
		? `<${returnType}>result.${abiFunction.outputs[0].name || 'result'}`
		: `<${returnType}>result`
	return `	public ${abiFunction.name}_ = async (${params}${separator}${attachedEthInputParameter}): Promise<${returnType}> => {
		const methodSignature = '${methodSignature}' as const
		const methodParameters = [${argNames}] as const
		const outputParameterDescriptions = ${JSON.stringify(abiFunction.outputs)} as const
		${abiFunction.outputs.length !== 0 ? 'const result = ' : ''}await this.localCall(methodSignature, outputParameterDescriptions, methodParameters${attachedEthCallParameter})${abiFunction.outputs.length !== 0 ? `\n\t\treturn ${returnValue}` : ''}
	}`
}

function toTsReturnTypeString(abiParameters: Array<AbiParameter>, errorContext: { contractName: string, functionName: string }): string {
	if (abiParameters.length === 0) return `void`
	else if (abiParameters.length === 1) return toTsTypeString(abiParameters[0], false, errorContext)
	else if (!abiParameters.every(abiParameter => !!abiParameter.name)) throw new Error(`Function ${errorContext.contractName}.${errorContext.functionName} has multiple return values but not all are named.`)
	else return `{${abiParameters.map(abiParameter => `${abiParameter.name}: ${toTsTypeString(abiParameter, false, errorContext)}`).join(', ')}}`
}

function toTsTypeString(abiParameter: AbiParameter, readonly: boolean, errorContext: { contractName: string, functionName: string }): string {
	if (/^u?int(\d*)$/.test(abiParameter.type)) {
		return `bigint`
	} else if (/^address$/.test(abiParameter.type)) {
		return 'Address'
	} else if (/^bool$/.test(abiParameter.type)) {
		return 'boolean'
	} else if (/^bytes\d+$/.test(abiParameter.type)) {
		const size = Number.parseInt(/^bytes(\d+)$/.exec(abiParameter.type)![1])
		return `Bytes${size}`
	} else if (/^string$/.test(abiParameter.type)) {
		return 'string'
	} else if (/^bytes$/.test(abiParameter.type)) {
		return 'Bytes'
	} else if (/^.*?\[\d*\]$/.test(abiParameter.type)) {
		const nestedAbiParameter = Object.assign({}, abiParameter, { type: /^(.*?)\[\d*\]$/.exec(abiParameter.type)![1] })
		return `${readonly ? 'Readonly' : ''}Array<${toTsTypeString(nestedAbiParameter, readonly, errorContext)}>`
	} else if (/^tuple$/.test(abiParameter.type)) {
		return `{ ${abiParameter.components!.map(component => `${readonly ? 'readonly ' : ''}${component.name}: ${toTsTypeString(component, readonly, errorContext)}`).join(', ')} }`
	} else if (/^u?fixed\d+x\d+$/.test(abiParameter.type)) {
		throw new Error(`fixed point parameters are not supported`)
	} else if (/^function$/.test(abiParameter.type)) {
		throw new Error(`function parameters are not supported`)
	} else {
		throw new Error(`Unrecognized parameter type in ${errorContext.contractName}.${errorContext.functionName}: ${JSON.stringify(abiParameter)}`)
	}
}

function toArgNameString(abiFunction: AbiFunction) {
	return abiFunction.inputs.map(toParamNameString).join(', ')
}

function toParamsString(abiFunction: AbiFunction, readonly: boolean, errorContext: { contractName: string }) {
	if (abiFunction.inputs.length == 0) return ''
	return abiFunction.inputs.map((abiParameter, i) => `${toParamNameString(abiParameter, i)}: ${toTsTypeString(abiParameter, readonly, { contractName: errorContext.contractName, functionName: abiFunction.name })}`).join(', ')
}

function toParamNameString(abiParameter: AbiParameter, index: number) {
	if (!abiParameter.name) return `arg${index}`
	else if (abiParameter.name.startsWith('_')) return abiParameter.name.substr(1)
	else return abiParameter.name
}

function toSignature(name: string, params: Array<AbiParameter>): string {
	const parameters = stringifyParamsForSignature(params).join(',')
	return `${name}(${parameters})`
}

function stringifyParamsForSignature(params: Array<AbiParameter>): Array<string> {
	return params.map(param => {
		if (param.type === 'tuple') {
			if (!param.components) throw new Error(`Expected components when type is ${param.type}`)
			return `(${stringifyParamsForSignature(param.components).join(',')})`
		} else if (param.type === 'tuple[]') {
			if (!param.components) throw new Error(`Expected components when type is ${param.type}`)
			return `(${stringifyParamsForSignature(param.components).join(',')})[]`
		} else {
			return param.type
		}
	})
}

// https://github.com/microsoft/TypeScript/issues/31535
interface TextEncoder {
	/** Returns "utf-8". */
	readonly encoding: string
	/** Returns the result of running UTF-8's encoder. */
	encode(input?: string): Uint8Array
}
declare var TextEncoder: { prototype: TextEncoder; new(): TextEncoder }
